@@ -135,14 +135,18 @@ _GET_AWS_CREDENTIALS_GQL = gql(
 
 _GET_BULK_DATA_STORAGE_GQL = gql(
     """
-    query getBulkDataStorage($tenant: String!) {
-        GetBulkDataStorage(tenant: $tenant, contentEncoding: gzip, count: 20) {
+    query getBulkDataStorage($tenant: String!, $useAccelerationEndpoint: Boolean!) {
+        GetBulkDataStorage(tenant: $tenant, count: 20, useAccelerationEndpoint: $useAccelerationEndpoint) {
+            expiration
             presignedGet
             presignedPost {
-                expiration
                 fields
                 url
             }
+            presignedPut {
+                headers
+                url
+            }
         }
     }
     """
@@ -265,6 +269,8 @@ class _NodeBotocoreSession(BotocoreSession):
 Auditor = Callable[..., dict[str, Any]]
 """Typing for MessageType auditor functions"""
 
+BatchItemFailures = dict[str, list[dict[str, str]]]
+
 
 @dataclass(frozen=True, init=False)
 class BulkDataStorage:
@@ -272,29 +278,40 @@ class BulkDataStorage:
     Class to manage bulk data storage.
     """
 
+    expiration: int
+    """Epoch, in seconds, when this expires"""
     presigned_get: str
     """URL that you can HTTP 'GET' to retrieve the bulk data"""
     presigned_post: PresignedPost
     """URL that you can HTTP 'POST' bulk data to, along with the fields the 'POST' requires"""
+    presigned_put: PresignedPut
+    """URL that you can HTTP 'PUT' bulk data to, along with the headers the 'PUT' requires"""
 
     def __init__(self, bulk_data_storage: dict[str, Union[str, PresignedPost]]) -> None:
         super().__init__()
+        super().__setattr__("expiration", bulk_data_storage["expiration"])
         super().__setattr__("presigned_get", bulk_data_storage["presignedGet"])
         super().__setattr__(
             "presigned_post",
             PresignedPost(
-                expiration=bulk_data_storage["presignedPost"]["expiration"],
                 fields=json.loads(bulk_data_storage["presignedPost"]["fields"]),
                 url=bulk_data_storage["presignedPost"]["url"],
             ),
         )
+        super().__setattr__(
+            "presigned_put",
+            PresignedPut(
+                headers=json.loads(bulk_data_storage["presignedPut"]["headers"]),
+                url=bulk_data_storage["presignedPut"]["url"],
+            ),
+        )
 
     @property
     def expired(self) -> bool:
         """
         Returns False if presigned_post is expired.
         """
-        return self.presigned_post.expiration < time()
+        return self.expiration < time()
 
 
 class CognitoRequestsHTTPTransport(RequestsHTTPTransport):
@@ -324,6 +341,20 @@ class Edge:
 LambdaEvent = Union[bool, dict, float, int, list, str, tuple, None]
 """Typing for the various types that a Lambda can be invoked with"""
 
+LambdaSqsRecords = list[
+    dict[
+        str,
+        Union[
+            str,
+            dict[str, str],
+            dict[
+                str,
+                dict[str, dict[str, Union[str, bytes, list[str], list[bytes]]]],
+            ],
+        ],
+    ]
+]
+
 
 @dataclass(frozen=True, init=False)
 class Message:
@@ -422,6 +453,7 @@ class Node(ABC):
         self,
         *,
         appsync_endpoint: str = None,
+        bulk_data_acceleration: bool = False,
         client_id: str = None,
         gql_transport_cls: type = CognitoRequestsHTTPTransport,
         name: str = None,
@@ -454,6 +486,7 @@ class Node(ABC):
             )["GetNode"]
         self.__app = data["app"]["name"]
         self.__app_type = data["app"]["__typename"]
+        self.__bulk_data_acceleration = bulk_data_acceleration
         self.__config: dict[str, Any] = None
         self.__gql_client = GqlClient(
             fetch_schema_from_transport=True,
@@ -465,18 +498,20 @@ class Node(ABC):
         self.__name = name
         self.__node_type = data["__typename"]
         self.__session = Session(
-            botocore_session=_NodeBotocoreSession(node=self, gql_client=gql_client)
+            botocore_session=_NodeBotocoreSession(node=self, gql_client=gql_client),
+            region_name=data["tenant"]["region"],
         )
         self.__sources: frozenset[Edge] = None
         self.__sqs_client: SQSClient = (
-            Session() if self.__app_type == "CrossAccountApp" else self.__session
+            Session(region_name=data["tenant"]["region"])
+            if self.__app_type == "CrossAccountApp"
+            else self.__session
         ).client(
             "sqs",
             config=Config(
                 max_pool_connections=min(20, ((cpu_count() or 1) + 4) * 2),
                 retries={"mode": "standard"},
             ),
-            region_name=data["tenant"]["region"],
         )
         self.__table: str = (
             data["tenant"]["table"] if data["app"].get("tableAccess") else None
@@ -519,6 +554,10 @@ class Node(ABC):
     def app_type(self) -> str:
         return self.__app_type
 
+    @property
+    def bulk_data_acceleration(self) -> bool:
+        return self.__bulk_data_acceleration
+
     @property
     def config(self) -> dict[str, Any]:
         return self.__config
@@ -534,6 +573,7 @@ class Node(ABC):
         *,
         group_id: str = None,
         previous_tracking_ids: Union[list[str], str] = None,
+        tracking_id: str,
     ) -> Message:
         """
         Creates message as per the message standard of echostream.
@@ -542,12 +582,14 @@ class Node(ABC):
         body - [POSITIONAL ARGUMENT] content of the message
         group_id - [KEYWORD ARGUMENT] group id
         previous_tracking_ids - [KEYWORD ARGUMENT] previous tracking id of the message if available
+        tracking_id - [KEYWORD ARGUMENT] tracking id of the message if available
         """
         return Message(
             body=body,
             group_id=group_id,
             message_type=self.send_message_type,
             previous_tracking_ids=previous_tracking_ids,
+            tracking_id=tracking_id,
         )
 
     @property
@@ -600,9 +642,20 @@ class PresignedPost:
     and are used to POST bulk data.
     """
 
-    expiration: int
-    """Epoch, in seconds, when this expires"""
     fields: dict[str, str]
     """The fields required to be sent when POSTing bulk data"""
     url: str
     """The POST url used to POST bulk data"""
+
+
+@dataclass(frozen=True)
+class PresignedPut:
+    """
+    PresignedPut objects are part of the Bulk Data Storage system
+    and are used to PUT bulk data.
+    """
+
+    headers: dict[str, str]
+    """The headers required to be sent when PUTing bulk data"""
+    url: str
+    """The PUT url used to PUT bulk data"""
