@@ -7,18 +7,23 @@ from functools import partial
 from gzip import GzipFile
 from io import BytesIO
 from queue import Empty, Queue
+from signal import SIGTERM, signal
 from threading import Event, RLock, Thread
 from time import sleep
 from typing import TYPE_CHECKING, Any, BinaryIO, Generator, Union
-from uuid import uuid4
 
 import dynamic_function_loader
 from aws_error_utils import catch_aws_error
 from requests import post
 
-from .. import _CREATE_AUDIT_RECORDS, _GET_BULK_DATA_STORAGE_GQL, _GET_NODE_GQL
+from .. import (
+    _CREATE_AUDIT_RECORDS,
+    _GET_BULK_DATA_STORAGE_GQL,
+    _GET_NODE_GQL,
+    BatchItemFailures,
+)
 from .. import BulkDataStorage as BaseBulkDataStorage
-from .. import Edge, LambdaEvent, Message, MessageType
+from .. import Edge, LambdaEvent, LambdaSqsRecords, Message, MessageType
 from .. import Node as BaseNode
 from .. import getLogger
 
@@ -99,7 +104,10 @@ class _BulkDataStorageQueue(Queue):
                         with node._gql_client as session:
                             bulk_data_storages: list[dict] = session.execute(
                                 _GET_BULK_DATA_STORAGE_GQL,
-                                variable_values={"tenant": node.tenant},
+                                variable_values={
+                                    "tenant": node.tenant,
+                                    "useAccelerationEndpoint": node.bulk_data_acceleration,
+                                },
                             )["GetBulkDataStorage"]
                 except Exception:
                     getLogger().exception("Error getting bulk data storage")
@@ -194,6 +202,7 @@ class Node(BaseNode):
         self,
         *,
         appsync_endpoint: str = None,
+        bulk_data_acceleration: bool = False,
         client_id: str = None,
         name: str = None,
         password: str = None,
@@ -204,6 +213,7 @@ class Node(BaseNode):
     ) -> None:
         super().__init__(
             appsync_endpoint=appsync_endpoint,
+            bulk_data_acceleration=bulk_data_acceleration,
             client_id=client_id,
             name=name,
             password=password,
@@ -252,6 +262,27 @@ class Node(BaseNode):
         except KeyError:
             raise ValueError(f"Unrecognized message type {message_type.name}")
 
+    def audit_messages(
+        self,
+        /,
+        messages: list[Message],
+        *,
+        extra_attributes: list[dict[str, Any]] = None,
+        source: str = None,
+    ) -> None:
+        """
+        Audits the provided messages. If extra_attibutes is
+        supplied they will be added to the respective message's audit
+        dict and they must have the same count as messages.
+        If source is provided, it will be recorded in the audit.
+        """
+        if extra_attributes and len(extra_attributes) != len(messages):
+            raise ValueError(
+                "messages and extra_attributes must have the same number of items"
+            )
+        for message, attributes in zip(messages, extra_attributes):
+            self.audit_message(message, extra_attributes=attributes, source=source)
+
     def handle_bulk_data(self, data: Union[bytearray, bytes]) -> str:
         """
         Posts data as bulk data and returns a GET URL for data retrieval.
@@ -357,7 +388,7 @@ class Node(BaseNode):
 
 
 class _DeleteMessageQueue(Queue):
-    def __init__(self, edge: Edge, node: Node) -> None:
+    def __init__(self, edge: Edge, node: AppNode) -> None:
         super().__init__()
 
         def deleter() -> None:
@@ -442,7 +473,7 @@ class _SourceMessageReceiver(Thread):
                         continue
                     getLogger().info(f"Received {len(sqs_messages)} from {edge.name}")
 
-                    handle_received_messages = [
+                    message_handlers = [
                         partial(
                             handle_received_message,
                             Message(
@@ -461,14 +492,23 @@ class _SourceMessageReceiver(Thread):
                         for sqs_message in sqs_messages
                     ]
 
-                    if executor := node._executor:
-                        wait(
-                            [executor.submit(func) for func in handle_received_messages]
-                        )
-                    else:
-                        for func in handle_received_messages:
-                            if not func():
-                                break
+                    def handle_received_messages() -> None:
+                        if executor := node._executor:
+                            wait(
+                                [
+                                    executor.submit(message_handler)
+                                    for message_handler in message_handlers
+                                ]
+                            )
+                        else:
+                            for message_handler in message_handlers:
+                                if not message_handler():
+                                    break
+
+                    Thread(
+                        name="handle_received_messages",
+                        target=handle_received_messages,
+                    ).start()
 
             getLogger().info(f"Stopping receiving messages from {edge.name}")
 
@@ -493,6 +533,7 @@ class AppNode(Node):
         self,
         *,
         appsync_endpoint: str = None,
+        bulk_data_acceleration: bool = False,
         client_id: str = None,
         executor: Executor = None,
         name: str = None,
@@ -504,6 +545,7 @@ class AppNode(Node):
     ) -> None:
         super().__init__(
             appsync_endpoint=appsync_endpoint,
+            bulk_data_acceleration=bulk_data_acceleration,
             client_id=client_id,
             name=name,
             password=password,
@@ -564,6 +606,7 @@ class LambdaNode(Node):
         self,
         *,
         appsync_endpoint: str = None,
+        bulk_data_acceleration: bool = False,
         client_id: str = None,
         concurrent_processing: bool = False,
         name: str = None,
@@ -576,15 +619,17 @@ class LambdaNode(Node):
     ) -> None:
         super().__init__(
             appsync_endpoint=appsync_endpoint,
+            bulk_data_acceleration=bulk_data_acceleration,
             client_id=client_id,
             name=name,
             password=password,
             tenant=tenant,
-            timeout=timeout,
+            timeout=timeout or 0.01,
             user_pool_id=user_pool_id,
             username=username,
         )
         self.start()
+        signal(SIGTERM, self._shutdown_handler)
         self.__executor: Executor = (
             ThreadPoolExecutor() if concurrent_processing else None
         )
@@ -596,7 +641,12 @@ class LambdaNode(Node):
     def _get_source(self, queue_arn: str) -> str:
         return self.__queue_name_to_source[queue_arn.split(":")[-1:][0]]
 
-    def handle_event(self, event: LambdaEvent) -> None:
+    def _shutdown_handler(self, signum: int, _: object) -> None:
+        getLogger().info("Received SIGTERM, shutting down")
+        self.join()
+        getLogger().info("Shutdown complete")
+
+    def handle_event(self, event: LambdaEvent) -> BatchItemFailures:
         """
         Handles the AWS Lambda event passed into the containing
         AWS Lambda function during invocation.
@@ -604,19 +654,7 @@ class LambdaNode(Node):
         This is intended to be the only called method in your
         containing AWS Lambda function.
         """
-        records: list[
-            dict[
-                str,
-                Union[
-                    str,
-                    dict[str, str],
-                    dict[
-                        str,
-                        dict[str, dict[str, Union[str, bytes, list[str], list[bytes]]]],
-                    ],
-                ],
-            ]
-        ] = None
+        records: LambdaSqsRecords = None
         if not (records := event.get("Records")):
             getLogger().warning(f"No Records found in event {event}")
             return
@@ -640,7 +678,7 @@ class LambdaNode(Node):
                 if self.__report_batch_item_failures:
                     batch_item_failures.remove(message_id)
 
-        handle_received_messages = [
+        message_handlers = [
             partial(
                 handle_received_message,
                 Message(
@@ -652,26 +690,26 @@ class LambdaNode(Node):
                     .get("stringValue"),
                     tracking_id=record["messageAttributes"]
                     .get("trackingId", {})
-                    .get("stringValue")
-                    or uuid4().hex,
+                    .get("stringValue"),
                 ),
                 record["messageId"],
             )
             for record in records
         ]
 
-        try:
-            if executor := self.__executor:
-                for future in as_completed(
-                    [executor.submit(func) for func in handle_received_messages]
-                ):
-                    if exception := future.exception():
-                        raise exception
-            else:
-                for func in handle_received_messages:
-                    func()
-        finally:
-            self.join()
+        if executor := self.__executor:
+            for future in as_completed(
+                [
+                    executor.submit(message_handler)
+                    for message_handler in message_handlers
+                ]
+            ):
+                if exception := future.exception():
+                    raise exception
+        else:
+            for message_handler in message_handlers:
+                message_handler()
+        self.join()
         if self.__report_batch_item_failures and batch_item_failures:
             return dict(
                 batchItemFailures=[
